@@ -3,6 +3,7 @@ Format and color output for tron commands.
 """
 from operator import itemgetter
 import os
+import sys
 
 
 class Color(object):
@@ -75,6 +76,8 @@ class TableDisplay(object):
         self.num_cols = self.console_width()
 
     def console_width(self):
+        if not sys.stdout.isatty():
+            return 80
         return int(os.popen('stty size', 'r').read().split()[1])
 
     def banner(self):
